use uefi::{
    prelude::BootServices,
    proto::{device_path::DevicePath, loaded_image::LoadedImage},
    table::boot::ScopedProtocol,
    CString16, Handle,
};

use crate::{
    device_path::{DevicePathBuf, DevicePathBufBuilder},
    Result,
};

pub struct Context<'a> {
    handle: Handle,
    image: ScopedProtocol<'a, LoadedImage>,
    boot_services: &'a BootServices,
}

impl<'a> Context<'a> {
    pub fn new(handle: Handle, boot_services: &'a BootServices) -> Result<Self> {
        let image = boot_services.open_protocol_exclusive::<LoadedImage>(handle)?;

        Ok(Self {
            handle,
            boot_services,
            image,
        })
    }

    pub fn handle(&self) -> Handle {
        self.handle
    }

    pub fn image(&self) -> &ScopedProtocol<LoadedImage> {
        &self.image
    }

    pub fn boot_services(&self) -> &BootServices {
        self.boot_services
    }

    pub fn file_path(&self) -> Option<&DevicePath> {
        self.image().file_path()
    }

    pub fn device_path(&self) -> Result<DevicePathBuf> {
        let device_path = self
            .boot_services()
            .open_protocol_exclusive::<DevicePath>(self.image().device())?;

        Ok(DevicePathBufBuilder::new()
            .with_device_path(&*device_path)?
            .finalize()?)
    }

    pub fn load_driver(&self, path: &str) -> Result<Handle> {
        let driver_path = CString16::try_from(path)?;

        let driver_device_path = DevicePathBufBuilder::new()
            .with_device_path(self.device_path().expect("device path"))
            .and_then(|builder| builder.with_file_path(&driver_path))
            .and_then(|builder| builder.finalize())
            .expect("device path");

        let handle = self
            .boot_services()
            .load_image(
                self.handle(),
                uefi::table::boot::LoadImageSource::FromFilePath {
                    file_path: &driver_device_path,
                    from_boot_manager: false,
                },
            )
            .expect("btrfs_driver_load");

        self.boot_services()
            .start_image(handle)
            .expect("start btrfs driver");

        Ok(handle)
    }
}
