use core::cell::{Ref, RefCell, RefMut};

use alloc::{boxed::Box, sync::Arc, vec};
use uefi::{
    proto::media::{
        file::{Directory as UefiDirectory, File, FileAttribute, FileHandle, FileInfo, FileMode},
        fs::SimpleFileSystem,
    },
    table::boot::ScopedProtocol,
    CStr16,
};

use crate::error::Result;

pub struct FileSystem<'a> {
    inner: RefCell<ScopedProtocol<'a, SimpleFileSystem>>,
}

pub mod dir {
    use core::ops::Deref;

    use alloc::boxed::Box;
    use uefi::proto::media::file::{File, FileAttribute, FileHandle, FileInfo, FileMode};

    use crate::Result;

    use super::Directory;

    pub struct DirectoryEntry {
        dir: Directory,
        file_info: Box<FileInfo>,
    }

    impl core::fmt::Debug for DirectoryEntry {
        fn fmt(&self, f: &mut core::fmt::Formatter<'_>) -> core::fmt::Result {
            f.debug_struct("DirectoryEntry")
                .field("file_info", &self.file_info)
                .finish()
        }
    }

    impl DirectoryEntry {
        pub fn open(
            &self,
            open_mode: FileMode,
            attribs: Option<FileAttribute>,
        ) -> Result<FileHandle> {
            let handle = self.dir.get_mut().open(
                self.file_info.file_name(),
                open_mode,
                attribs.unwrap_or(FileAttribute::empty()),
            )?;

            Ok(handle)
        }
    }

    impl AsRef<FileInfo> for DirectoryEntry {
        fn as_ref(&self) -> &FileInfo {
            &self.file_info
        }
    }

    impl Deref for DirectoryEntry {
        type Target = FileInfo;

        fn deref(&self) -> &Self::Target {
            &self.file_info
        }
    }

    impl Directory {
        pub fn next_entry(&mut self) -> Result<Option<Box<FileInfo>>> {
            Ok(self.get_entry()?)
        }
    }

    impl Iterator for Directory {
        type Item = DirectoryEntry;

        fn next(&mut self) -> Option<Self::Item> {
            self.next_entry()
                .ok()
                .flatten()
                .map(|file_info| DirectoryEntry {
                    dir: self.clone(),
                    file_info,
                })
        }
    }
}

#[repr(transparent)]
#[derive(Clone)]
pub struct Directory {
    dir: Arc<RefCell<UefiDirectory>>,
}

impl Directory {
    pub fn get(&self) -> Ref<'_, UefiDirectory> {
        RefCell::borrow(&self.dir)
    }

    pub fn get_mut<'a>(&'a self) -> RefMut<'a, UefiDirectory> {
        RefCell::borrow_mut(&self.dir)
    }

    pub fn open(
        &self,
        filename: &CStr16,
        open_mode: FileMode,
        attributes: FileAttribute,
    ) -> uefi::Result<FileHandle> {
        self.get_mut().open(filename, open_mode, attributes)
    }

    pub fn get_entry(&mut self) -> uefi::Result<Option<Box<FileInfo>>> {
        self.get_mut().read_entry_boxed()
    }
}

impl From<UefiDirectory> for Directory {
    fn from(value: UefiDirectory) -> Self {
        Self {
            dir: Arc::new(RefCell::new(value)),
        }
    }
}

impl<'a> FileSystem<'a> {
    pub fn new(inner: ScopedProtocol<'a, SimpleFileSystem>) -> Self {
        Self {
            inner: RefCell::new(inner),
        }
    }

    pub fn open(
        &self,
        path: &str,
        open_mode: FileMode,
        attribs: Option<FileAttribute>,
    ) -> Result<FileHandle> {
        let mut buf = vec![0; path.len() + 1];
        let cstr = CStr16::from_str_with_buf(path, &mut buf)?;
        Ok(self.root_dir()?.get_mut().open(
            cstr,
            open_mode,
            attribs.unwrap_or(FileAttribute::empty()),
        )?)
    }

    pub fn root_dir(&self) -> Result<Directory> {
        Ok(self.inner.borrow_mut().open_volume()?.into())
    }
}
