use alloc::{borrow::ToOwned, vec::Vec};
use itertools::Itertools;
use terminal::image::{EncodableLayout, RgbaImage};
use uefi::{proto::console::gop, table::boot::ScopedProtocol};

use crate::{error::Error, Result};

pub struct GraphicsOutput<'a, 'b> {
    gop: ScopedProtocol<'a, gop::GraphicsOutput<'b>>,
    info: gop::ModeInfo,
    #[allow(dead_code)]
    backbuffer: Vec<gop::BltPixel>,
}

impl<'a, 'b> GraphicsOutput<'a, 'b> {
    pub fn from_gop(mut gop: ScopedProtocol<'a, gop::GraphicsOutput<'b>>) -> Result<Self> {
        let mode = {
            let mode = gop
                .modes()
                .map(|mode| {
                    let info = mode.info().to_owned();
                    (mode, info.resolution().0 * info.resolution().1)
                })
                .sorted_by(|a, b| a.1.cmp(&b.1))
                .next();

            mode.map(|(mode, _)| mode)
        }
        .ok_or(Error::NoModeFound)?;

        gop.set_mode(&mode)?;

        let info = mode.info();
        //let pixel_width =

        Ok(Self::new(gop, *info))
    }

    pub fn resolution(&self) -> (usize, usize) {
        self.info.resolution()
    }

    pub fn from_gop_with_mode(
        mut gop: ScopedProtocol<'a, gop::GraphicsOutput<'b>>,
        mode: gop::Mode,
    ) -> Result<Self> {
        gop.set_mode(&mode)?;

        Ok(Self::new(gop, *mode.info()))
    }

    pub fn new(gop: ScopedProtocol<'a, gop::GraphicsOutput<'b>>, info: gop::ModeInfo) -> Self {
        let black = gop::BltPixel::new(0, 0, 0);
        Self {
            gop,
            info,
            backbuffer: alloc::vec![black; info.stride() * info.resolution().1],
        }
    }

    fn image_as_bltpixels(image: &RgbaImage) -> &[gop::BltPixel] {
        let len = (image.width() * image.height()) as usize;

        unsafe {
            core::slice::from_raw_parts(image.as_bytes().as_ptr() as *const gop::BltPixel, len)
        }
    }

    pub fn blt_image(&mut self, image: &RgbaImage) -> uefi::Result<()> {
        let (width, height) = image.dimensions();
        let op = gop::BltOp::BufferToVideo {
            buffer: Self::image_as_bltpixels(image),
            src: gop::BltRegion::Full,
            dest: (0, 0),
            dims: (width as usize, height as usize),
        };
        self.gop.blt(op)?;
        Ok(())
    }

    pub fn blt_image_position(
        &mut self,
        x: usize,
        y: usize,
        image: &RgbaImage,
    ) -> uefi::Result<()> {
        let op = gop::BltOp::BufferToVideo {
            buffer: Self::image_as_bltpixels(image),
            src: gop::BltRegion::Full,
            dest: (x, y),
            dims: (image.width() as usize, image.height() as usize),
        };
        self.gop.blt(op)?;
        Ok(())
    }
}
