#![feature(
    allocator_api,
    ptr_metadata,
    error_in_core,
    alloc_error_handler,
    negative_impls,
    int_roundings,
    core_intrinsics
)]
#![no_std]
#![no_main]

use alloc::{collections::BTreeMap, vec};
use log::info;
use serde::Deserialize;
use uefi::{
    prelude::*,
    proto::{
        console::gop,
        device_path::DevicePath,
        loaded_image::LoadedImage,
        media::{
            file::{File, FileAttribute, FileInfo, FileMode},
            fs::SimpleFileSystem,
        },
    },
    table::{
        boot::{OpenProtocolAttributes, OpenProtocolParams, SearchType},
        Boot, SystemTable,
    },
    CString16, Identify, Status,
};

extern crate alloc;

use bootloader::{
    device_path::{DevicePathBuf, DevicePathBufBuilder},
    fpu::enable_fpu,
    fs,
    graphics::GraphicsOutput,
    Result,
};

#[allow(dead_code)]
#[derive(Debug, Deserialize)]
pub struct Config {
    timeout: Option<u64>,
}

#[derive(Debug)]
pub struct Context {
    handle: Handle,
    st: SystemTable<Boot>,
}

impl Context {
    pub fn new(handle: Handle, mut st: SystemTable<Boot>) -> uefi::Result<Self> {
        uefi_services::init(&mut st).expect("services");
        enable_fpu().expect("no FPU/SSE/SSE2!");
        Ok(Self { handle, st })
    }

    pub fn device_path(&self) -> Result<DevicePathBuf> {
        let loaded_image = self
            .boot_services()
            .open_protocol_exclusive::<LoadedImage>(self.handle())?;
        let device_path = self
            .boot_services()
            .open_protocol_exclusive::<DevicePath>(loaded_image.device())?;

        Ok(DevicePathBufBuilder::new()
            .with_device_path(&*device_path)?
            .finalize()?)
    }

    pub fn fs(&self) -> Result<fs::FileSystem> {
        let file_system = self.st.boot_services().get_image_file_system(self.handle)?;

        Ok(fs::FileSystem::new(file_system))
    }

    pub fn system_table(&self) -> &SystemTable<Boot> {
        &self.st
    }

    pub fn handle(&self) -> Handle {
        self.handle
    }

    pub fn system_table_mut(&mut self) -> &mut SystemTable<Boot> {
        &mut self.st
    }

    pub fn boot_services(&self) -> &BootServices {
        self.st.boot_services()
    }

    pub fn get_graphics_output(&self) -> Result<GraphicsOutput> {
        let handles = self
            .boot_services()
            .locate_handle_buffer(SearchType::ByProtocol(&gop::GraphicsOutput::GUID))?;

        let gops = handles.handles().iter().filter_map(|handle| unsafe {
            self.boot_services()
                .open_protocol::<gop::GraphicsOutput>(
                    OpenProtocolParams {
                        handle: handle.clone(),
                        agent: self.handle,
                        controller: None,
                    },
                    OpenProtocolAttributes::GetProtocol,
                )
                .ok()
        });

        let (_, (gop, mode)) = gops
            .map(|gop| {
                let (pixel_count, mode) = gop
                    .modes()
                    .map(|mode| {
                        let (x, y) = mode.info().resolution();
                        (x * y, mode)
                    })
                    .collect::<BTreeMap<_, _>>()
                    .pop_last()
                    .expect("gop has 0 modes !?");

                (pixel_count, (gop, mode))
            })
            .collect::<BTreeMap<_, _>>()
            .pop_last()
            .expect("no GOP found!");

        GraphicsOutput::from_gop_with_mode(gop, mode)
    }

    pub fn load_driver(&self, path: &str) -> Result<Handle> {
        let driver_path = CString16::try_from(path)?;

        let driver_device_path = DevicePathBufBuilder::new()
            .with_device_path(self.device_path().expect("device path"))
            .and_then(|builder| builder.with_file_path(&driver_path))
            .and_then(|builder| builder.finalize())
            .expect("device path");

        let handle = self
            .boot_services()
            .load_image(
                self.handle(),
                uefi::table::boot::LoadImageSource::FromFilePath {
                    file_path: &driver_device_path,
                    from_boot_manager: false,
                },
            )
            .expect("btrfs_driver_load");

        self.boot_services()
            .start_image(handle)
            .expect("start btrfs driver");

        Ok(handle)
    }

    pub fn run(mut self) -> Result<Status> {
        info!("Hello, UEFI!");
        self.system_table_mut().stdin().reset(false)?;

        let fs = self.fs()?;

        let filename = CString16::try_from("config.toml").unwrap();
        let mut config_file =
            fs.root_dir()?
                .open(&filename, FileMode::Read, FileAttribute::empty())?;

        let mut buffer = vec![0; config_file.get_boxed_info::<FileInfo>()?.file_size() as usize];
        config_file
            .into_regular_file()
            .unwrap()
            .read(&mut buffer)
            .unwrap();

        let config = toml::from_str::<Config>(&core::str::from_utf8(&buffer).expect("from_utf8"))
            .expect("toml parse");

        info!("config: {:?}", config);

        drop(fs);

        self.load_driver("bootloader\\drivers\\btrfs_x64.efi")
            .expect("driver");

        let handles = self
            .boot_services()
            .find_handles::<SimpleFileSystem>()
            .expect("SFS handles");

        for handle in handles {
            if let Ok(mut sfs) = self
                .boot_services()
                .open_protocol_exclusive::<SimpleFileSystem>(handle)
            {
                let mut dir = sfs.open_volume().expect("root dir");
                let info = dir.get_boxed_info::<FileInfo>();

                info!("{:?}", info);
            };
        }

        info!("done");
        loop {}
    }
}

#[entry]
fn efi_entry(handle: Handle, system_table: SystemTable<Boot>) -> Status {
    let ctx = Context::new(handle, system_table).unwrap();
    ctx.run().unwrap()
}
