//#![allow(dead_code)]
#![feature(error_in_core)]
#![cfg_attr(not(any(feature = "std", test)), no_std)]

use core::{
    cell::{RefCell, RefMut},
    mem::size_of,
};

use alloc::{
    boxed::Box,
    collections::{btree_map::Entry, BTreeMap},
    vec,
    vec::Vec,
};
use scroll::Pread;
use thiserror::Error;

use structs::{Chunk, Key, KeyPtr, KnownObjectId, ObjectType, RootItem, Stripe, Superblock};
use tree::Tree;

extern crate alloc;

pub mod crc32c;
pub mod files;
pub mod path;
pub mod structs;
pub mod tree;
pub mod v2;

#[cfg(feature = "std")]
pub mod std_io {
    use std::io::{Cursor, Read, Seek};

    use crate::{Error, VolumeIo};

    impl<T: Read + Seek> VolumeIo for T {
        fn read(&mut self, dst: &mut [u8], address: u64) -> Result<(), Error> {
            self.seek(std::io::SeekFrom::Start(address))
                .map_err(|a| Error::ReadFailed)?;
            self.read_exact(dst).map_err(|_| Error::ReadFailed)
        }

        fn alignment(&self) -> usize {
            1
        }
    }
}

pub use v2::error::*;

pub trait VolumeIo {
    fn read(&mut self, dst: &mut [u8], address: u64) -> Result<()>;
    fn alignment(&self) -> usize;
}

/// equal if overlapping, ordered by lower bound
#[derive(Debug)]
pub struct ChunkTreeKey {
    range: core::ops::Range<u64>,
}

impl From<u64> for ChunkTreeKey {
    fn from(value: u64) -> Self {
        Self {
            range: value..value,
        }
    }
}

impl ChunkTreeKey {
    pub fn start(&self) -> u64 {
        self.range.start
    }
    pub fn end(&self) -> u64 {
        self.range.end
    }
    pub fn size(&self) -> u64 {
        self.range.end - self.range.start
    }

    pub fn delta(&self, point: u64) -> u64 {
        point - self.range.start
    }

    pub fn sub_range(&self, point: u64) -> core::ops::Range<u64> {
        self.delta(point)..self.end()
    }
}

impl Eq for ChunkTreeKey {}
impl Ord for ChunkTreeKey {
    fn cmp(&self, other: &Self) -> core::cmp::Ordering {
        self.partial_cmp(other).unwrap()
    }
}

impl PartialEq for ChunkTreeKey {
    fn eq(&self, other: &Self) -> bool {
        self.range.contains(&other.range.start) || other.range.contains(&self.range.start)
    }
}

impl PartialOrd for ChunkTreeKey {
    fn partial_cmp(&self, other: &Self) -> Option<core::cmp::Ordering> {
        self.eq(other)
            .then_some(core::cmp::Ordering::Equal)
            .or_else(|| self.range.start.partial_cmp(&other.range.start))
    }
}

type ChunkTree = BTreeMap<ChunkTreeKey, u64>;

#[derive(Debug)]
pub struct Btrfs<R: VolumeIo> {
    reader: Box<RefCell<R>>,
    superblock: Superblock,
    pub chunk_cache: ChunkTree,
    roots: BTreeMap<KnownObjectId, RootItem>,
}

#[derive(Debug)]
pub struct Root<'a, R: VolumeIo> {
    id: KnownObjectId,
    tree: Tree<'a, R>,
}

#[derive(Debug)]
pub struct Volume<'a, R: VolumeIo> {
    fs: &'a Btrfs<R>,
    root: Root<'a, R>,
}

impl<R: VolumeIo> Btrfs<R> {
    pub fn reader(&self) -> RefMut<R> {
        self.reader.borrow_mut()
    }

    pub fn new(mut reader: R) -> Result<Self> {
        let mut buf = vec![0; size_of::<Superblock>()];
        reader.read(&mut buf, Superblock::SUPERBLOCK_BASE_OFFSET as _)?;
        let superblock = Superblock::parse(&buf)?;
        let chunk_cache = Self::bootstrap_chunk_tree(&superblock)?;
        let mut new = Self {
            reader: Box::new(RefCell::new(reader)),
            superblock,
            chunk_cache,
            roots: Default::default(),
        };

        new.parse_chunk_tree()?;
        new.parse_root_tree()?;

        Ok(new)
    }

    pub fn size_from_logical(&self, logical: u64) -> Option<u64> {
        self.chunk_cache
            .get_key_value(&logical.into())
            .map(|(key, _)| key.size())
    }

    pub fn offset_from_logical(&self, logical: u64) -> Option<u64> {
        self.chunk_cache
            .get_key_value(&logical.into())
            .map(|(key, offset)| offset + key.delta(logical))
    }

    pub fn range_from_logical(&self, logical: u64) -> Option<core::ops::Range<u64>> {
        self.chunk_cache
            .get_key_value(&logical.into())
            .map(|(key, offset)| {
                let delta = key.delta(logical);
                (offset + delta)..(offset + key.size() - delta)
            })
    }

    pub fn read_range(&self, range: core::ops::Range<u64>) -> Result<Vec<u8>> {
        let mut buf = vec![0; (range.end - range.start) as usize];
        self.reader().read(&mut buf, range.start)?;

        Ok(buf)
    }

    fn parse_root_tree(&mut self) -> Result<()> {
        let root_tree = Tree::from_logical_offset(self, self.superblock().root.get())?;

        let roots = root_tree
            .full_range()
            .filter_map(|(item, v)| match v {
                structs::TreeItem::Root(root) => Some((item, root)),
                _ => None,
            })
            .map(|(item, root)| {
                let id = item.key.id();

                (id, root)
            })
            .collect::<Vec<_>>();

        self.roots.extend(roots);

        Ok(())
    }

    fn parse_chunk_tree(&mut self) -> Result<()> {
        let chunk_tree = Tree::from_logical_offset(self, self.superblock().chunk_root.get())?;

        let chunks = chunk_tree
            .full_range()
            .filter_map(|(item, v)| match v {
                structs::TreeItem::Chunk(chunk) => Some((item, chunk)),
                _ => None,
            })
            .collect::<Vec<_>>();

        for (item, chunk) in chunks {
            let start = item.key.offset.get() as u64;
            let end = start + chunk.length.get();

            match self.chunk_cache.entry(ChunkTreeKey { range: start..end }) {
                Entry::Vacant(entry) => {
                    log::info!("inserting chunk [{start}, {end})");
                    entry.insert(chunk.stripe.offset.get());
                }
                Entry::Occupied(entry) => {
                    log::warn!("overlapping stripes!");
                    log::warn!(
                        "\t{:?} and {:?}",
                        entry.key(),
                        ChunkTreeKey { range: start..end }
                    );
                    log::warn!(
                        "\twith offsets: {} and {}",
                        entry.get(),
                        chunk.stripe.offset.get()
                    );

                    if *entry.get() != chunk.stripe.offset.get() {
                        log::error!("\tprobably an error?");
                    }
                }
            }
        }

        Ok(())
    }

    pub fn read_keyptr(&self, keyptr: &KeyPtr) -> Result<Vec<u8>> {
        self.read_range(
            self.range_from_logical(keyptr.blockptr.get())
                .ok_or(Error::ReadFailed)?,
        )
    }

    pub fn bootstrap_chunk_tree(superblock: &Superblock) -> Result<ChunkTree> {
        let array_size = superblock.sys_chunk_array_size.get() as usize;
        let mut offset: usize = 0;

        let key_size = size_of::<Key>();
        let mut chunk_tree = ChunkTree::new();

        let bytes = &superblock.sys_chunk_array;

        while offset < array_size {
            if offset + key_size > array_size {
                log::error!("short key read");
                return Err(Error::InvalidOffset);
            }

            let key = bytes.gread::<Key>(&mut offset)?;
            if key.ty.as_type() != ObjectType::ChunkItem {
                log::error!("key is not of type ChunkItem");
                return Err(Error::InvalidOffset);
            }

            let chunk = bytes.gread::<Chunk>(&mut offset)?;
            let num_stripes = chunk.num_stripes.get(); // copy to prevent unaligned access

            if num_stripes == 0 {
                log::error!("num_stripes cannot be 0");
                return Err(Error::InvalidOffset);
            }

            if num_stripes != 1 {
                log::warn!(
                    "warning: {} stripes detected but only processing 1",
                    num_stripes
                );
            }

            let key_offset = key.offset.get();
            let chunk_length = chunk.length.get();

            match chunk_tree.entry(ChunkTreeKey {
                range: key_offset..(key_offset + chunk_length),
            }) {
                Entry::Vacant(entry) => {
                    entry.insert(chunk.stripe.offset.get());
                }
                Entry::Occupied(_) => {
                    log::error!("overlapping stripes!");
                    return Err(Error::InvalidOffset);
                }
            };

            offset += (num_stripes - 1) as usize * size_of::<Stripe>();
            if offset > array_size {
                log::error!("short chunk item + stripes read");
                return Err(Error::InvalidOffset);
            }
        }

        Ok(chunk_tree)
    }

    pub fn iter_roots(&self) {
        let root_tree =
            Tree::from_logical_offset(self, self.superblock().root.get()).expect("tree");

        let range = root_tree.full_range();

        for (item, value) in range {
            if item.key.ty.as_type() == ObjectType::RootItem {
                let root = value.as_root().expect("root");
                let tree = Tree::from_logical_offset(self, root.bytenr.get()).expect("tree");

                log::info!("some tree idek which one innit");
                for (i, v) in tree.full_range() {
                    log::info!("{i:?}");
                    log::info!("{v:?}");
                }
            }
        }
    }

    pub fn find_default_subvol(&self) -> Result<Option<Volume<'_, R>>> {
        let root_tree = Tree::from_logical_offset(self, self.superblock().root.get())?;

        let _range = root_tree.full_range();

        // we are looking for the root tree directory (?)
        // this is a DIR_ITEM entry in the root tree, with the name "default",
        // and the crc32 of "default" as its offset
        let key = Key::new(
            KnownObjectId::Custom(self.superblock.root_dir_objectid.get()),
            ObjectType::DirItem,
            0x8dbfc2d2, // crc of "default"
        );

        if let Some(subvol_root) = root_tree.find_key(&key)? {
            // if we found the dir entry of the "default subvol" (mharmstone nomenclature)
            // we then look for the root fs tree in the root tree with the ID found in the `.location` of the dir_item only (from mharmstone)

            let subvol_id = subvol_root
                .1
                .as_dir_item()
                .expect("dir item")
                .first()
                .expect("dir item entry")
                .item()
                .location
                .id();

            if let Some(root) = self.roots.get(&subvol_id) {
                log::info!("my root: {root:#?}");

                let root = Root {
                    id: subvol_id,
                    tree: Tree::from_logical_offset(self, root.bytenr.get())?,
                };

                Ok(Some(Volume { fs: self, root }))
            } else {
                Ok(None)
            }
        } else {
            log::warn!("default subvol not found");
            Ok(None)
        }
    }

    pub fn superblock(&self) -> &Superblock {
        &self.superblock
    }
}
